import type_enforced
from fizgrid.utils import unique_id, RectangleMoverUtils


@type_enforced.Enforcer(enabled=True)
class Entity:
    def __init__(
        self,
        name: str,
        shape: list[list[int | float]],
        x_coord: int | float,
        y_coord: int | float,
    ):
        """
        Initializes an entity with a given shape and location in the grid.

        Args:

        - id (int): The ID of the entity.
        - name (str): The name of the entity.
        - shape (list[list[int|float]]): The shape of the entity as a list of points centered around the shape origin.
            - The shape origin referenced here should be the center of the shape as the shape origin is used to determine how the shape is located on the grid.
            - The shape is a list of points, where each point is a list of two coordinates [x, y] relative to the shape origin.
        - x_coord (int|float): The starting x-coordinate of the entity in the grid.
        - y_coord (int|float): The starting y-coordinate of the entity in the grid.
        - grid (Grid): The grid the entity is in.
        """
        self.id = unique_id()
        """The ID of the entity."""
        self.name = name
        """The name of the entity."""
        self.shape = shape
        """The shape of the entity as a list of points centered around the shape the current (x, y) coordinates."""
        self.x_coord = x_coord
        """
        The x-coordinate of the entity in the grid. This is the last known location of the entity.
        
        - Note: This is only updated when events realize the route. This means that a route in progress would not have the correct
            location until the route is realized (either completed or interrupted by a cancellation or collision).
        """
        self.y_coord = y_coord
        """
        The y-coordinate of the entity in the grid. This is the last known location of the entity.
        
        - Note: This is only updated when events realize the route. This means that a route in progress would not have the correct
            location until the route is realized (either completed or interrupted by a cancellation or collision).
        """
        self.history = []
        """
        The history of the entity's location and collision status. 
        This is a list of dictionaries containing the x, y, t, and c values.

        - x (int|float): The x-coordinate of the entity in the grid.
        - y (int|float): The y-coordinate of the entity in the grid.
        - t (int|float): The time at which the entity was at this location.
        - c (bool): Whether the entity was in a collision at this time.

        - Note: This is only updated when events realize the route. This means that a route in progress would not have history until the
            route is realized (either completed or interrupted by a cancellation or collision).
        """

        # Util Attributes
        self.__grid__ = None
        self.__on_grid__ = False
        self.__route_start_time__ = None
        self.__route_end_time__ = 0
        self.__blocked_grid_cells__ = []
        self.__planned_waypoints__ = []
        self.__future_event_ids__ = {}

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name})"

    def __assoc_grid__(self, grid) -> None:
        """
        Associate the grid to this entity.
        Creates a reference in both the grid and the entity to allow easy access from both directions.

        Args:

        - grid (Grid): The grid to assign to this entity.
        """
        if self.__grid__ is not None:
            raise Exception(
                f"Entity {self.name} is already associated with a grid. Cannot associate with a new grid."
            )
        self.__grid__ = grid

    def __place_on_grid__(self) -> None:
        """
        Place this entity on the grid and claim the grid cells it will block.

        This allows the entity to take up space and move around the grid.

        Args:

        - grid (Grid): The grid to assign to this entity.
        """
        if self.__grid__ is None:
            raise Exception(
                f"Entity {self.name} is not associated with a grid. Cannot place on grid."
            )
        self.__on_grid__ = True
        self.__route_start_time__ = self.get_time()
        self.history.append(
            {
                "x": self.x_coord,
                "y": self.y_coord,
                "t": self.get_time(),
                "c": False,
            }
        )
        self.__realize_route__(
            is_result_of_collision=False,
            raise_on_future_collision=True,
            is_result_of_dissoc_grid=False,
        )

    def __dissoc_grid__(self) -> None:
        """
        Dissociate the grid from this entity.
        This method removes the entity from the grid and clears any blocked grid cells and future events.
        """
        if self.__grid__ is None:
            raise Exception(
                f"Entity {self.name} is not associated with a grid. Cannot dissociate from grid."
            )
        if self.__on_grid__:
            self.__realize_route__(
                is_result_of_collision=False,
                raise_on_future_collision=False,
                is_result_of_dissoc_grid=True,
            )
        # Clear the blocked grid cells and future events
        self.__clear_blocked_grid_cells__()
        self.__clear_future_events__()
        self.__route_start_time__ = None
        self.__grid__ = None
        self.__on_grid__ = False

    def __clear_blocked_grid_cells__(self) -> None:
        """
        Clears the blocked grid cells for this entity.
        """
        for x_cell, y_cell, block_id in self.__blocked_grid_cells__:
            cell = self.__grid__.__cells__[y_cell][x_cell]
            cell.pop(block_id, None)
        self.__blocked_grid_cells__ = []

    def __clear_future_events__(self) -> None:
        """
        Clears the future events for this entity.
        """
        for (
            this_event_id,
            related_event_id,
        ) in self.__future_event_ids__.items():
            # Remove the event from the queue
            event_obj = self.__grid__.__queue__.remove_event(this_event_id)
            # If this event is a standard route_end event, it will not have an related event, so we can be done here
            # If the event_obj is None, it has already been processed or removed so we can skip any further processing
            # If this event is a collision event, it will have an associated event and should be removed too
            if related_event_id != None and event_obj != None:
                self.__grid__.__queue__.remove_event(related_event_id)
        self.__future_event_ids__ = {}

    def __waypoint_check__(self, waypoints) -> None:
        """
        Checks the waypoints for validity.
        Raises an exception if the waypoints are not valid.

        Args:

        - waypoints (list[tuple[int|float,int|float,int|float]]): A list of waypoints to be added to the grid queue.
        """
        for waypoint in waypoints:
            if len(waypoint) != 3:
                raise Exception(
                    f"Waypoint must be a tuple of (x_coord, y_coord, time_shift). Waypoint: {waypoint}"
                )
            if waypoint[0] < 0 or waypoint[1] < 0 or waypoint[2] < 0:
                raise Exception(
                    f"Waypoint coordinates and times must be positive. Waypoint: {waypoint}"
                )
            if (
                waypoint[0] > self.__grid__.__x_size__
                or waypoint[1] > self.__grid__.__y_size__
            ):
                raise Exception(
                    f"Waypoint coordinates must be within the grid. Waypoint: {waypoint}"
                )

    def __plan_route__(
        self,
        waypoints: list[tuple[int | float, int | float, int | float]],
        raise_on_future_collision: bool = False,
    ) -> dict:
        """
        Sets the route for this entity given a set of waypoints starting at the current time.
        Determines the cells this entity will block and at which times it will block them.
        Checks for collisions with other entities.
        Adds events to the queue for the first collision with each colliding entity.
          - Note: Not all events will occur, and future events will be cleared when the first event occurs.

        Args:

        - waypoints (list[tuple[int|float,int|float,int|float]]): A list of waypoints to be added to the grid queue.
            - A list of tuples where each tuple is (x_coord, y_coord, time_shift).
            - EG;
                ```
                waypoints = [
                    (5, 3, 10),
                    (3, 5, 10)
                ]
                ```
                - Move to (5, 3) over 10 seconds
                - Move to (3, 5) over 10 seconds
            - Note: x_coord and y_coord are the coordinates of the waypoint. They must both be positive.
            - Note: time_shift is the time it takes to move to the waypoint. It must be positive.
        - raise_on_future_collision (bool): Whether to raise an exception if the entity has any future plans that result in a collision.
            - Note: This will raise an exception even if the future collision would not happen due to another event occurring first.
            - Note: This is mostly used when initially placing entities on the grid to ensure they are not placed in a collision.

        Returns:

        - dict: A dictionary containing the following keys:
            - has_collision (bool): Whether the route has a collision with another entity.
        """
        if not self.__on_grid__:
            raise Exception(
                "Entity is not on this grid yet, but is attempting to plan a route. Either it has not been assigned to this grid, or the time it will be placed on the grid is in the future."
            )
        # Raise an exception if the entity is already in a route
        if not self.is_available():
            raise Exception(
                f"entity {self.name} is not available for a new route. Cannot set a new route until the current route is finished."
            )

        # Check for valid waypoints
        self.__waypoint_check__(waypoints)

        # Setup util attributes
        self.__clear_blocked_grid_cells__()
        self.__clear_future_events__()

        x_tmp = self.x_coord
        y_tmp = self.y_coord
        t_tmp = self.get_time()

        collisions = {}
        total_route_time_shift = sum([waypoint[2] for waypoint in waypoints])

        # Add a final waypoint occuring until the end of the simulation.
        # This allows us to lock in the position of the entity at the end of the route and block the grid cells accordingly.
        if len(waypoints) > 0:
            waypoints.append(
                (
                    waypoints[-1][0],
                    waypoints[-1][1],
                    self.__grid__.__max_time__ - t_tmp - total_route_time_shift,
                )
            )
        else:
            waypoints.append(
                (
                    self.x_coord,
                    self.y_coord,
                    self.__grid__.__max_time__ - t_tmp - total_route_time_shift,
                )
            )

        # Store the route waypoints and start time for later use to determine the entity's position at a given time
        self.__planned_waypoints__ = waypoints
        self.__route_start_time__ = self.get_time()
        self.__route_end_time__ = min(
            self.__grid__.__max_time__,
            self.__route_start_time__ + total_route_time_shift,
        )

        # For each route waypoint, calculate the blocks and collisions and add them to the grid
        for waypoint in waypoints:
            blocks = RectangleMoverUtils.moving_shape_overlap_intervals(
                x_coord=x_tmp,
                y_coord=y_tmp,
                x_shift=waypoint[0] - x_tmp,
                y_shift=waypoint[1] - y_tmp,
                t_start=t_tmp,
                t_end=t_tmp + waypoint[2],
                shape=self.shape,
            )
            x_tmp = waypoint[0]
            y_tmp = waypoint[1]
            t_tmp = t_tmp + waypoint[2]
            for (x_cell, y_cell), (t_start, t_end) in blocks.items():
                # Check if the cell is within the grid bounds
                if (
                    x_cell < 0
                    or y_cell < 0
                    or x_cell >= self.__grid__.__x_size__
                    or y_cell >= self.__grid__.__y_size__
                ):
                    # Note: Shape coords outside of the grid raise an exception, but this this would indicate that the shape may have an overlap over the edge.
                    # Note: If there are exterior walls, no exception here is necessary
                    # TODO: Determine if this should be an exception or just a warning
                    # print(f"Warning: Entity {self.name} is outside of the grid bounds is outside of the grid bounds based on its shape.")
                    continue
                # Store a unique block_id to allow for removal of the block later
                block_id = unique_id()
                # Get the relevant cell in the grid
                cell = self.__grid__.__cells__[y_cell][x_cell]
                # Check for collisions with other entities in the cell
                for (
                    other_t_start,
                    other_t_end,
                    other_entity_id,
                ) in cell.values():
                    if t_start < other_t_end and t_end > other_t_start:
                        # Determine the time of the collision and store the most recent collision time with each colliding entity
                        collision_time = max(t_start, other_t_start)
                        previous_collision_time = collisions.get(
                            other_entity_id
                        )
                        if (
                            previous_collision_time is None
                            or collision_time < previous_collision_time
                        ):
                            collisions[other_entity_id] = collision_time
                # Block the grid cell for the entity
                cell[block_id] = (t_start, t_end, self.id)
                # Store the blocked grid cell for later removal
                self.__blocked_grid_cells__.append((x_cell, y_cell, block_id))
        if raise_on_future_collision and len(collisions) > 0:
            raise Exception(
                f"{self.__repr__()} collides with other entities now or in the future. "
            )
        # Create collision events for the first collision with each colliding entity
        for other_entity_id, collision_time in collisions.items():
            other_entity = self.__grid__.__entities__[other_entity_id]
            event_id = self.__grid__.add_event(
                time=collision_time,
                object=self,
                method="__realize_route__",
                kwargs={
                    "is_result_of_collision": True,
                },
            )
            other_event_id = self.__grid__.add_event(
                time=collision_time,
                object=other_entity,
                method="__realize_route__",
                kwargs={
                    "is_result_of_collision": True,
                },
            )
            # Store the event_id for each entity involved in the collision
            self.__future_event_ids__[event_id] = other_event_id
            other_entity.__future_event_ids__[other_event_id] = event_id

        if self.__route_end_time__ > self.get_time():
            # Add a route_end event for this entity at the timing of the end of the route
            event_id = self.__grid__.add_event(
                time=self.__route_end_time__,
                object=self,
                method="__realize_route__",
                kwargs={
                    "is_result_of_collision": False,
                },
            )
            self.__future_event_ids__[event_id] = None
        return {
            "has_collision": len(collisions) > 0,
        }

    def __realize_route__(
        self,
        is_result_of_collision: bool = False,
        raise_on_future_collision: bool = False,
        is_result_of_dissoc_grid: bool = False,
    ) -> dict:
        """
        Realize the route for this entity at the current time.

        Args:

        - is_result_of_collision (bool): Whether this route end is the result of a collision.
        - raise_on_future_collision (bool): Whether to raise an exception if the entity is in a future collision.
            - Raises an exception if this event causes a future collision with another entity.
        - is_result_of_dissoc_grid (bool): Whether the entity will be removed from the grid after this event.

        Returns:

        - dict: A dictionary containing the following keys:
            - has_collision (bool): Whether the route has a collision with another entity.
        """
        # Determeine Realized Route and update the entity's position / history
        x_tmp = self.x_coord
        y_tmp = self.y_coord
        t_tmp = self.__route_start_time__
        current_time = self.get_time()
        for waypoint in self.__planned_waypoints__:
            # End the route realization if the time is greater than the current time
            if t_tmp >= current_time:
                if t_tmp > current_time:
                    # TODO: Remove this
                    print("This should never happen.")
                break
            # Get partial location if interrupted by the current time
            elif t_tmp + waypoint[2] > current_time:
                pct_complete = (current_time - t_tmp) / waypoint[2]
                x_tmp = (waypoint[0] - x_tmp) * pct_complete + x_tmp
                y_tmp = (waypoint[1] - y_tmp) * pct_complete + y_tmp
                t_tmp = current_time
            # Otherwise, update the tmp location
            else:
                x_tmp = waypoint[0]
                y_tmp = waypoint[1]
                t_tmp = t_tmp + waypoint[2]

            self.history.append(
                {
                    "x": x_tmp,
                    "y": y_tmp,
                    "t": t_tmp,
                    "c": is_result_of_collision,
                }
            )

        # Set the entity's position to where they are at this point in time
        self.x_coord = x_tmp
        self.y_coord = y_tmp
        # Set the entity's route end time to the current time
        self.__route_end_time__ = current_time

        if is_result_of_dissoc_grid:
            return {"is_result_of_collision": False}
        # Stop the entity at their current location and update the grid for their expected future
        planned_route = self.__plan_route__(
            waypoints=[],
            raise_on_future_collision=raise_on_future_collision,
        )
        self.on_realize(is_result_of_collision=is_result_of_collision)
        return planned_route

    def get_time(self) -> int | float:
        """
        Returns the current time of the grid queue.
        This method retrieves the current time from the grid queue.

        Returns:

        - int|float: The current time of the grid queue.
        """
        return self.__grid__.__queue__.__time__

    def is_available(self) -> bool:
        """
        Returns whether this entity is available for a new route.

        Returns:

        - bool: True if the entity is available for a new route, False otherwise.
        """
        return self.__route_end_time__ <= self.get_time()

    def add_route(
        self,
        waypoints: list[tuple[int | float, int | float, int | float]],
        time: int | float | None = None,
    ) -> None:
        """
        Adds a route to the grid for this entity.

        Args:

        - waypoints (list[tuple[int|float,int|float,int|float]]): A list of waypoints to be added to the grid queue.
            - A list of tuples where each tuple is (x_coord, y_coord, time_shift).
            - EG:
                ```
                waypoints = [
                    (5, 3, 10),
                    (3, 5, 10)
                ]
                ```
                - Move to (5, 3) over 10 seconds
                - Move to (3, 5) over 10 seconds
            - Note: x_coord and y_coord are the coordinates of the waypoint. They must both be positive.
            - Note: time_shift is the time it takes to move to the waypoint. It must be positive.
        - time (int|float|None): The time at which to start the route. If None, the current time is used.
        """
        if self.__grid__ is None:
            raise Exception(
                "Entity is not assigned to a grid. Cannot add a route."
            )
        if time is None:
            time = self.get_time()
        # Add the event to the queue
        self.__grid__.add_event(
            time=time,
            object=self,
            method="__plan_route__",
            kwargs={"waypoints": waypoints},
        )

    def cancel_route(
        self,
        time: int | float | None = None,
    ) -> None:
        """
        Cancels the route for this entity.

        Args:

        - time (int|float|None): The time at which to cancel the route. If None, the current time is used.
        """
        if self.__grid__ is None:
            raise Exception(
                "Entity is not assigned to a grid. Cannot cancel a route."
            )
        if time is None:
            time = self.get_time()
        # Add the event to the queue
        self.__grid__.add_event(
            time=time,
            object=self,
            method="__realize_route__",
            kwargs={},
        )

    def on_realize(self, **kwargs):
        """
        Called when the route is realized to allow for custom behavior.
        This method should be overridden by subclasses to implement custom behavior.

        Args:

        - **kwargs: Additional arguments passed to the method. Subclasses may use this or add their own arguments.
        """


class StaticEntity(Entity):
    """
    This class is used to represent entities that do not move, such as walls or other static objects.

    This is an extension of the Entity class that represents a static entity.

    To improve efficiency, many events and logic are avoided as a static entity does not move or respond to events.
    """

    def __realize_route__(
        self,
        is_result_of_collision: bool = False,
        raise_on_future_collision: bool = False,
        is_result_of_dissoc_grid: bool = False,
    ) -> dict:
        """
        Realize the route for this entity at the current time.

        Args:

        - is_result_of_collision (bool): Whether this route end is the result of a collision.
            - If True, the route end is the result of a collision and the entity should not be allowed to start a new route until the collision is resolved.
            - If False, the route end is not the result of a collision and the entity should be allowed to start a new route.
        - raise_on_future_collision (bool): Whether to raise an exception if the entity is in a future collision.
            - Raises an exception if this event causes a future collision with another entity.
        - is_result_of_dissoc_grid (bool): Whether the entity will be removed from the grid after this event.

        Returns:

        - dict: A dictionary containing the following keys:
            - has_collision (bool): Whether the route has a collision with another entity.
        """
        # Since this is a static entity, we don't need to do anything here.
        if is_result_of_collision:
            return
        if is_result_of_dissoc_grid:
            super().__realize_route__(
                is_result_of_collision=False,
                raise_on_future_collision=False,
                is_result_of_dissoc_grid=True,
            )
        # Since this is a static entity, the route end time is the current time when the entity is created (should normally be 0)
        self.__route_end_time__ = self.get_time()
        # Since this object does not move, we don't need to plan a route and will never interrupt it.
        return self.__plan_route__(
            waypoints=[], raise_on_future_collision=raise_on_future_collision
        )


class GhostEntity(Entity):
    """
    This class is used to represent entities that do not collide with other entities.

    It is an extension of the Entity class that represents a ghost entity.

    Essentially, these entities can move freely without worrying about collisions. Using them avoids substantial overhead in the simulation and can dramatically improve performance if collisions are irrelevant.
    """

    def __plan_route__(
        self,
        waypoints: list[tuple[int | float, int | float, int | float]],
        raise_on_future_collision: bool = False,
    ) -> dict:
        """
        Overrides the __plan_route__ method to allow for a ghost entity to be placed on the grid that never collides with other entities because it does not check for collisions.

        Sets the route for this entity given a set of waypoints starting at the current time.

        Args:

        - waypoints (list[tuple[int|float,int|float,int|float]]): A list of waypoints to be added to the grid queue.
            - A list of tuples where each tuple is (x_coord, y_coord, time_shift).
            - EG;
                ```
                waypoints = [
                    (5, 3, 10),
                    (3, 5, 10)
                ]
                ```
                - Move to (5, 3) over 10 seconds
                - Move to (3, 5) over 10 seconds
            - Note: x_coord and y_coord are the coordinates of the waypoint. They must both be positive.
            - Note: time_shift is the time it takes to move to the waypoint. It must be positive.
        - raise_on_future_collision (bool): Whether to raise an exception if the entity has any future plans that result in a collision.
            - Note: This is not used for ghost entities, but is included for consistency with the parent class.

        Returns:

        - dict: A dictionary containing the following keys:
            - has_collision (bool): Whether the route has a collision with another entity.
                - Note: This will always be False for ghost entities.
        """
        # Raise an exception if the entity is already in a route
        if not self.is_available():
            raise Exception(
                f"entity {self.name} is not available for a new route. Cannot set a new route until the current route is finished."
            )

        # Check for valid waypoints
        self.__waypoint_check__(waypoints)

        # Setup util attributes
        self.__clear_future_events__()

        total_route_time_shift = sum([waypoint[2] for waypoint in waypoints])

        self.__route_start_time__ = self.get_time()

        # Add a final waypoint occuring until the end of the simulation.
        # This allows us to lock in the position of the entity at the end of the route and block the grid cells accordingly.
        if len(waypoints) > 0:
            waypoints.append(
                (
                    waypoints[-1][0],
                    waypoints[-1][1],
                    self.__grid__.__max_time__
                    - self.__route_start_time__
                    - total_route_time_shift,
                )
            )
        else:
            waypoints.append(
                (
                    self.x_coord,
                    self.y_coord,
                    self.__grid__.__max_time__
                    - self.__route_start_time__
                    - total_route_time_shift,
                )
            )

        # Store the route waypoints and start time for later use to determine the entity's position at a given time
        self.__planned_waypoints__ = waypoints
        self.__route_end_time__ = min(
            self.__grid__.__max_time__,
            self.__route_start_time__ + total_route_time_shift,
        )

        # Add a route_end event for this entity at the timing of the end of the route
        if self.__route_start_time__ < self.__route_end_time__:
            event_id = self.__grid__.add_event(
                time=self.__route_end_time__,
                object=self,
                method="__realize_route__",
                kwargs={
                    "is_result_of_collision": False,
                },
            )
            self.__future_event_ids__[event_id] = None
        return {"has_collision": False}



# Testing for TimeQueue class
from fizgrid.queue import TimeQueue

passing = True


class EventClass:
    def custom_event(self, string: str):
        print(string)


eventMaker = EventClass()


# Creating a TimeQueue instance
queue = TimeQueue()

# Creating Empty QueueEvent classes
event_0 = {"callable": eventMaker.custom_event, "string": "Event 0"}
event_1 = {"callable": eventMaker.custom_event, "string": "Event 1"}
event_2 = {"callable": eventMaker.custom_event, "string": "Event 2"}
event_3 = {"callable": eventMaker.custom_event, "string": "Event 3"}

# Populating some example events
t0_id = queue.add_event(time=5, event=event_0)
t1_id = queue.add_event(time=10, event=event_1)
t2_id = queue.add_event(time=7, event=event_2)
t3_id = queue.add_event(time=8, event=event_3)

# Removing an event
queue.remove_event(t3_id)

# Checking the order of events
if queue.get_next_event() != {"id": t0_id, "time": 5, "event": event_0}:
    passing = False
if queue.get_next_event() != {"id": t2_id, "time": 7, "event": event_2}:
    passing = False
if queue.get_next_event() != {"id": t1_id, "time": 10, "event": event_1}:
    passing = False
if queue.get_next_event() != {"id": None, "time": None, "event": None}:
    passing = False

if passing:
    print("test_02.py: passed")
else:
    print("test_02.py: failed")



from fizgrid.grid import Grid
from fizgrid.entities import Entity
from fizgrid.utils import Shape
import random, math


class Sniffer(Entity):
    def __init__(self, *args, **kwargs):
        # Override the init method of the Entity class to extend the functionality and add custom attributes.
        super().__init__(*args, **kwargs)
        self.goal_x = None
        self.goal_y = None
        self.tolerance = None
        self.speed = 1

    def get_dist_from_goal(self):
        # A function to calculate the distance from the entity to the goal.
        return (
            (self.goal_x - self.x_coord) ** 2
            + (self.goal_y - self.y_coord) ** 2
        ) ** 0.5

    def set_goal(self, x_coord, y_coord, tolerance=1):
        # Set the goal for the entity to reach and start the routing process.
        self.goal_x = x_coord
        self.goal_y = y_coord
        self.tolerance = tolerance
        self.add_next_route()

    def add_next_route(self):
        # Determine the next route for the entity to take.

        # This route is randomly calculated based on the angle and distance to the goal.
        #     - If the entity is within the tolerance of the goal, it will stop.
        #     - If the entity is not within the tolerance
        #         - It will calculate the target angle towards the goal.
        #         - It will calculate the target distance to the goal.
        #         - A random angle is generated by using a normal distribution centered on the goal angle.
        #         - A random distance is generated between 0 and the minimum of the distance to the goal and 5.
        distance_from_goal = self.get_dist_from_goal()
        if distance_from_goal < self.tolerance:
            return
        goal_angle_rad = math.atan2(
            self.goal_y - self.y_coord, self.goal_x - self.x_coord
        )
        random_angle = random.normalvariate(goal_angle_rad, math.pi / 2)
        distance = random.uniform(0, min(distance_from_goal, 5))
        x_shift = distance * math.cos(random_angle)
        y_shift = distance * math.sin(random_angle)

        self.add_route(
            waypoints=[
                (
                    self.x_coord + x_shift,
                    self.y_coord + y_shift,
                    distance * self.speed,
                ),
            ]
        )

    def on_realize(self, **kwargs):
        # This method is a blank method that should be overridden by the user.
        # It is called when an event for the entity is realized.

        # In this case, it is used to implment the goal checking logic and continuously add routes until the goal is reached.

        # The method checks if the entity has a goal set and if it is within the tolerance of the goal.
        # If the entity is not within the tolerance, it will add the next route.
        # If the entity is within the tolerance, it will set the goal to None.
        if self.goal_x is not None and self.goal_y is not None:
            if self.get_dist_from_goal() > self.tolerance:
                self.add_next_route()
            else:
                self.goal_x = None
                self.goal_y = None


forest = Grid(
    name="forest",
    x_size=1000,
    y_size=1000,
    add_exterior_walls=True,
)

# Add some pigs to the forest
truffle_pig_1 = forest.add_entity(
    Sniffer(
        name="Truffle_Pig_1",
        shape=Shape.rectangle(x_len=1, y_len=1, round_to=2),
        x_coord=450,
        y_coord=500,
    )
)
truffle_pig_2 = forest.add_entity(
    Sniffer(
        name="Truffle_Pig_2",
        shape=Shape.rectangle(x_len=1, y_len=1, round_to=2),
        x_coord=500,
        y_coord=450,
    )
)

truffle_pig_1.set_goal(
    x_coord=550,
    y_coord=500,
)

truffle_pig_2.set_goal(
    x_coord=500,
    y_coord=550,
)

# Run the sim
forest.simulate()

# print({
#     'Name': truffle_pig_1.name,
#     'x_coord': truffle_pig_1.x_coord,
#     'y_coord': truffle_pig_1.y_coord,
# })
# print({
#     'Name': truffle_pig_2.name,
#     'x_coord': truffle_pig_2.x_coord,
#     'y_coord': truffle_pig_2.y_coord,
# })

# Example Output
# {'Name': 'Truffle_Pig_1', 'x_coord': 550.6344767270211, 'y_coord': 499.90150976359365}
# {'Name': 'Truffle_Pig_2', 'x_coord': 499.2754299204255, 'y_coord': 550.6818320798665}


# Checks for success
success = True
truffle_pig_1.goal_x = 550
truffle_pig_1.goal_y = 500

truffle_pig_2.goal_x = 500
truffle_pig_2.goal_y = 550

if truffle_pig_1.get_dist_from_goal() > 1:
    success = False
if truffle_pig_2.get_dist_from_goal() > 1:
    success = False

if success:
    print("test_05.py: passed")
else:
    print("test_05.py: failed")



"""
# Fizgrid
[![PyPI version](https://badge.fury.io/py/fizgrid.svg)](https://badge.fury.io/py/fizgrid)
[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)

Simulate entities that take up space over time in grid based environments. 


# Overview
This package introduces a new approach to simulating physical movement within a grid. Rather than following the traditional method of discretizing time and iterating over spatial configurations, this simulation inverts the model: space is discretized, and time is continuous.

Events are scheduled into a priority queue, ordered by their occurrence time. The simulation processes events in chronological order, updating only the relevant entities and grid cells affected by each event. This design allows for highly efficient simulation of large numbers of entities over long periods of simulated time.

## Entity Movement and Scheduling

Entity movements are planned in advance but only realized when an associated event is triggered. When a route is assigned to an entity, each grid cell along its path is reserved for the time the entity is expected to occupy it. During this reservation process, the system checks for potential conflicts. For example, whether another entity is scheduled to occupy the same cell at the same time.

If a conflict is detected, a collision event is added to the queue for both entities. Only the first potential collision between any pair of entities is added to avoid redundancy. Additionally, an end-of-route event is queued for the entity to signal the completion of its movement.

## Event Handling

When an end-of-route event is processed:

- The entity’s position is updated to where it would be at that time.
- All pending events related to its previous route are removed from the queue.
- A placeholder (blank) route is assigned to the entity to occupy its current position until a new route is defined.
    - This may trigger new collisions, which will be resolved in their own time.

When a collision event is processed:

- The entity is moved to its calculated position at the collision time.
- All future events tied to its current route are removed.
- Collision events for other entities that would have involved this entity are also removed.
    - The counterpart entity in the current collision still processes its corresponding event.
- A blank route is assigned to the entity to mark its occupancy of the current cell until reassigned.
    - This may trigger new collisions will be resolved in their own time.

# Setup

Make sure you have Python 3.10.x (or higher) installed on your system. You can download it [here](https://www.python.org/downloads/).

## Installation

```
pip install fizgrid
```

## Basic Usage

### Technical Docs
The technical documentation is available at [https://connor-makowski.github.io/fizgrid/index.html](https://connor-makowski.github.io/fizgrid/index.html).

### Example
```py
from fizgrid.grid import Grid
from fizgrid.entities import Entity, StaticEntity, GhostEntity
from fizgrid.utils import Shape

# Create a grid with exterior walls
grid = Grid(
    name="living_room",
    x_size=10,
    y_size=10,
    add_exterior_walls=True,
)

# Add some static entities
sofa = grid.add_entity(
    StaticEntity(
        name="sofa",
        shape=Shape.rectangle(x_len=2, y_len=1),
        x_coord=5,
        y_coord=5.5,
    )
)

# Add a dynamic entity
robot = grid.add_entity(
    Entity(
        name="robot",
        shape=Shape.rectangle(x_len=1, y_len=1),
        x_coord=5,
        y_coord=2,
    )
)

# Add a dynamic entity that can pass through other entities
ghost = grid.add_entity(
    GhostEntity(
        name="ghost",
        shape=Shape.rectangle(x_len=1, y_len=1),
        x_coord=5,
        y_coord=2,
    )
)

# Attempt to move the robot through the sofa
# Starting at time=0, the robot will try to move to (5,8) over 12 seconds
# This will bump into the sofa and stop
robot.add_route(time=0, waypoints=[(5, 8, 12)])


# Attempt to move the ghost through the sofa
# Starting at time=0, the ghost will try to move to (5,8) over 12 seconds
# This will not bump into the sofa and complete its route
ghost.add_route(time=0, waypoints=[(5, 8, 12)])

# Simulate the grid
grid.simulate()

# Show the history of the robot
print(robot.history)
# [{'x': 5, 'y': 2, 't': 0, 'c': False}, {'x': 5.0, 'y': 4.5, 't': 5.0, 'c': True}]
# This means that the robot started at (5,2) at time=0 and moved to (5,4.5) at time=5 where it collided with the sofa
# Since there is no additional logic to handle the collision, the robot just stopped moving at time 5 and waited for the end of the simulation


# Show the history of the ghost
print(ghost.history)
# [{'x': 5, 'y': 2, 't': 0, 'c': False}, {'x': 5, 'y': 8, 't': 12, 'c': False}]
# This means that the ghost started at (5,2) at time=0 and moved to (5,8) at time=12 without any collisions
```

# More Comprehensive Examples

## Create a basic sniffer entity
Create a Truffle Pig that sniffs out truffles on a grid. 
```py
from fizgrid.grid import Grid
from fizgrid.entities import Entity
from fizgrid.utils import Shape
import random, math


class Sniffer(Entity):
    def __init__(self, *args, **kwargs):
        # Override the init method of the Entity class to extend the functionality and add custom attributes.
        super().__init__(*args, **kwargs)
        self.goal_x = None
        self.goal_y = None
        self.tolerance = None
        self.speed = 1

    def get_dist_from_goal(self):
        # A function to calculate the distance from the entity to the goal.
        return (
            (self.goal_x - self.x_coord) ** 2
            + (self.goal_y - self.y_coord) ** 2
        ) ** 0.5

    def set_goal(self, x_coord, y_coord, tolerance=1):
        # Set the goal for the entity to reach and start the routing process.
        self.goal_x = x_coord
        self.goal_y = y_coord
        self.tolerance = tolerance
        self.add_next_route()

    def add_next_route(self):
        # Determine the next route for the entity to take.

        # This route is randomly calculated based on the angle and distance to the goal.
        #     - If the entity is within the tolerance of the goal, it will stop.
        #     - If the entity is not within the tolerance
        #         - It will calculate the target angle towards the goal.
        #         - It will calculate the target distance to the goal.
        #         - A random angle is generated by using a normal distribution centered on the goal angle.
        #         - A random distance is generated between 0 and the minimum of the distance to the goal and 5.
        distance_from_goal = self.get_dist_from_goal()
        if distance_from_goal < self.tolerance:
            return
        goal_angle_rad = math.atan2(
            self.goal_y - self.y_coord, self.goal_x - self.x_coord
        )
        random_angle = random.normalvariate(goal_angle_rad, math.pi / 2)
        distance = random.uniform(0, min(distance_from_goal, 5))
        x_shift = distance * math.cos(random_angle)
        y_shift = distance * math.sin(random_angle)

        self.add_route(
            waypoints=[
                (
                    self.x_coord + x_shift,
                    self.y_coord + y_shift,
                    distance * self.speed,
                ),
            ]
        )

    def on_realize(self, **kwargs):
        # This method is a blank method that should be overridden by the user.
        # It is called when an event for the entity is realized.

        # In this case, it is used to implment the goal checking logic and continuously add routes until the goal is reached.

        # The method checks if the entity has a goal set and if it is within the tolerance of the goal.
        # If the entity is not within the tolerance, it will add the next route.
        # If the entity is within the tolerance, it will set the goal to None.
        if self.goal_x is not None and self.goal_y is not None:
            if self.get_dist_from_goal() > self.tolerance:
                self.add_next_route()
            else:
                self.goal_x = None
                self.goal_y = None


forest = Grid(
    name="forest",
    x_size=1000,
    y_size=1000,
    add_exterior_walls=True,
)

# Add some pigs to the forest
truffle_pig_1 = forest.add_entity(
    Sniffer(
        name="Truffle_Pig_1",
        shape=Shape.rectangle(x_len=1, y_len=1, round_to=2),
        x_coord=450,
        y_coord=500,
    )
)
truffle_pig_2 = forest.add_entity(
    Sniffer(
        name="Truffle_Pig_2",
        shape=Shape.rectangle(x_len=1, y_len=1, round_to=2),
        x_coord=500,
        y_coord=450,
    )
)

truffle_pig_1.set_goal(
    x_coord=550,
    y_coord=500,
)

truffle_pig_2.set_goal(
    x_coord=500,
    y_coord=550,
)

# Run the sim
forest.simulate()

print({
    'Name': truffle_pig_1.name,
    'x_coord': truffle_pig_1.x_coord,
    'y_coord': truffle_pig_1.y_coord,
})
print({
    'Name': truffle_pig_2.name,
    'x_coord': truffle_pig_2.x_coord,
    'y_coord': truffle_pig_2.y_coord,
})

# Example Output
# {'Name': 'Truffle_Pig_1', 'x_coord': 550.6344767270211, 'y_coord': 499.90150976359365}
# {'Name': 'Truffle_Pig_2', 'x_coord': 499.2754299204255, 'y_coord': 550.6818320798665}
```


# Development
## Running Tests, Prettifying Code, and Updating Docs

Make sure Docker is installed and running on a Unix system (Linux, MacOS, WSL2).

- Create a docker container and drop into a shell
    - `./run.sh`
- Run all tests (see ./utils/test.sh)
    - `./run.sh test`
- Prettify the code (see ./utils/prettify.sh)
    - `./run.sh prettify`
- Update the docs (see ./utils/docs.sh)
    - `./run.sh docs`

- Note: You can and should modify the `Dockerfile` to test different python versions.
"""



import math, type_enforced, uuid


def unique_id():
    """
    Generates a unique identifier.
    """
    return str(uuid.uuid4())


class Thunk(type_enforced.utils.Partial):
    def __call__(self, *args, **kwargs):
        new_args = self.__args__ + args
        new_kwargs = {**self.__kwargs__, **kwargs}
        self.__arity__ = self.__getArity__(new_args, new_kwargs)
        if self.__arity__ < 0:
            self.__exception__("Too many arguments were supplied")
        if self.__arity__ == 0 and len(args) == 0 and len(kwargs) == 0:
            results = self.__fn__(*new_args, **new_kwargs)
            return results
        return Thunk(
            self.__fn__,
            *new_args,
            **new_kwargs,
        )

    def __get__(self, instance, owner):
        def bind(*args, **kwargs):
            if instance is not None and self.__arity__ == self.__fnArity__:
                return self.__call__(instance, *args, **kwargs)
            else:
                return self.__call__(*args, **kwargs)

        return bind


@type_enforced.Enforcer(enabled=True)
class Shape:
    @staticmethod
    def circle(radius: int, points: int = 6, round_to: int = 2) -> list:
        """
        Returns a list of addative coordinates that form a circle around a given point.
        """
        return [
            [
                round(radius * math.cos(2 * math.pi / points * i), round_to),
                round(radius * math.sin(2 * math.pi / points * i), round_to),
            ]
            for i in range(points)
        ]

    @staticmethod
    def rectangle(
        x_len: float | int, y_len: float | int, round_to: int = 2
    ) -> list:
        """
        Returns a list of addative coordinates that form a rectangle around a given point.
        """
        return [
            [round(x_len / 2, round_to), round(y_len / 2, round_to)],
            [round(-x_len / 2, round_to), round(y_len / 2, round_to)],
            [round(-x_len / 2, round_to), round(-y_len / 2, round_to)],
            [round(x_len / 2, round_to), round(-y_len / 2, round_to)],
        ]


class RectangleMoverUtils:
    @staticmethod
    def moving_segment_overlap_intervals(
        seg_start: int | float,
        seg_end: int | float,
        t_start: int | float,
        t_end: int | float,
        shift: int | float,
    ):
        """
        Calculates the time intervals during which a moving 1D line segment overlaps with each unit-length
        integer-aligned range along the x-axis.

        Args:

            - seg_start (int|float): Initial position of the left end of the line segment.
            - seg_end (int|float): Initial position of the right end of the line segment.
            - t_start (int|float): Start time of the motion.
            - t_end (int|float): End time of the motion.
            - shift (int|float): Total distance the line segment moves along the x-axis during [t_start, t_end].

        Returns:

            - dict[int, tuplie(int|float,int|float)]: A dictionary mapping each integer `i` to the time interval [t_in, t_out]
                                    during which any part of the line overlaps the range [i, i+1).
                                    Only includes ranges with non-zero overlap duration.
        """
        duration = t_end - t_start
        velocity = shift / duration if duration != 0 else 0

        result = {}

        final_start = seg_start + shift
        final_end = seg_end + shift
        global_min = min(seg_start, final_start)
        global_max = max(seg_end, final_end)

        for i in range(int(global_min) - 1, int(global_max) + 2):
            if velocity == 0:
                if seg_end > i and seg_start < i + 1 and t_start < t_end:
                    result[i] = (t_start, t_end)
                continue
            # Solve for times when the line enters and exits overlap with [i, i+1)
            t1 = (i - seg_end) / velocity + t_start
            t2 = (i + 1 - seg_start) / velocity + t_start
            entry_time = max(min(t1, t2), t_start)
            exit_time = min(max(t1, t2), t_end)

            if exit_time > entry_time:
                result[i] = (entry_time, exit_time)

        return result

    @staticmethod
    def moving_rectangle_overlap_intervals(
        x_start: float | int,
        x_end: float | int,
        y_start: float | int,
        y_end: float | int,
        x_shift: float | int,
        y_shift: float | int,
        t_start: float | int,
        t_end: float | int,
    ):
        """
        Calculates the time intervals during which a moving rectangle overlaps with each unit-length
        integer-aligned range along the x and y axes.

        Args:

            - x_start (float|int): Initial position of the left end of the rectangle along the x-axis.
            - x_end (float|int): Initial position of the right end of the rectangle along the x-axis.
            - y_start (float|int): Initial position of the bottom end of the rectangle along the y-axis.
            - y_end (float|int): Initial position of the top end of the rectangle along the y-axis.
            - x_shift (float|int): Total distance the rectangle moves along the x-axis during [t_start, t_end].
            - y_shift (float|int): Total distance the rectangle moves along the y-axis during [t_start, t_end].
            - t_start (float|int): Start time of the motion.
            - t_end (float|int): End time of the motion.

        Returns:

            - dict[tuple(int,int),tuple(int|float,int|float)]: A dictionary mapping each integer (i,j) to the time interval [t_in, t_out]
                during which any part of the rectangle overlaps the range [i, i+1) x [j, j+1).
                Only includes ranges with non-zero overlap duration.

        """
        x_intervals = RectangleMoverUtils.moving_segment_overlap_intervals(
            seg_start=x_start,
            seg_end=x_end,
            t_start=t_start,
            t_end=t_end,
            shift=x_shift,
        )
        y_intervals = RectangleMoverUtils.moving_segment_overlap_intervals(
            seg_start=y_start,
            seg_end=y_end,
            t_start=t_start,
            t_end=t_end,
            shift=y_shift,
        )
        result = {}
        for x_key, x_interval in x_intervals.items():
            for y_key, y_interval in y_intervals.items():
                # Only add intervals with time overlap
                if (
                    x_interval[1] > y_interval[0]
                    and y_interval[1] > x_interval[0]
                ):
                    result[(x_key, y_key)] = (
                        max(x_interval[0], y_interval[0]),
                        min(x_interval[1], y_interval[1]),
                    )

        return result

    @staticmethod
    def moving_shape_overlap_intervals(
        x_coord: float | int,
        y_coord: float | int,
        x_shift: float | int,
        y_shift: float | int,
        t_start: float | int,
        t_end: float | int,
        shape: list[list[float | int]],
    ):
        """
        Calculates the time intervals during which a moving shape overlaps with each unit-length
        integer-aligned range along the x and y axes.

        Note: This converts each shape into a full bounding box rectangle and then uses the rectangle overlap function to calculate the intervals.

        Args:

            - x_coord (float|int): Initial x-coordinate of the shape's center.
            - y_coord (float|int): Initial y-coordinate of the shape's center.
            - x_shift (float|int): Total distance the shape moves along the x-axis during [t_start, t_end].
            - y_shift (float|int): Total distance the shape moves along the y-axis during [t_start, t_end].
            - t_start (float|int): Start time of the motion.
            - t_end (float|int): End time of the motion.
            - shape (list[list[float|int]]): List of coordinates representing the shape's vertices relative to its center.

        Returns:

            - dict[tuple(int,int),tuple(int|float,int|float)]: A dictionary mapping each integer (i,j) to the time interval [t_in, t_out]
                                    during which any part of the shape overlaps the range [i, i+1) x [j, j+1).
                                    Only includes ranges with non-zero overlap duration.
        """
        # Return the overlap intervals for a rectangle
        return RectangleMoverUtils.moving_rectangle_overlap_intervals(
            x_start=min([x_coord + coord[0] for coord in shape]),
            x_end=max([x_coord + coord[0] for coord in shape]),
            y_start=min([y_coord + coord[1] for coord in shape]),
            y_end=max([y_coord + coord[1] for coord in shape]),
            x_shift=x_shift,
            y_shift=y_shift,
            t_start=t_start,
            t_end=t_end,
        )



import type_enforced
from type_enforced.utils import WithSubclasses
from fizgrid.entities import Entity, StaticEntity
from fizgrid.queue import TimeQueue


@type_enforced.Enforcer(enabled=True)
class Grid:
    def __init__(
        self,
        name: str,
        x_size: int,
        y_size: int,
        max_time: int = 1000,
        add_exterior_walls: bool = True,
    ):
        """
        Initializes a grid with the specified parameters.

        Args:

        - name (str): The name of the grid.
        - x_size (int): The width of the grid.
        - y_size (int): The height of the grid.
        - max_time (int): The maximum time for the grid simulation.
        - add_exterior_walls (bool): Whether to add exterior walls to the grid.
            - Default: True
        """
        # Passed Attributes
        self.name: str = name
        """The name of the grid."""
        self.__x_size__ = x_size
        self.__y_size__ = y_size
        self.__max_time__ = max_time

        # Calculated Attributes
        self.__entities__ = {}
        self.__queue__ = TimeQueue()
        self.__cells__ = [[{} for _ in range(x_size)] for _ in range(y_size)]

        if add_exterior_walls:
            self.add_exterior_walls()

    def __repr__(self):
        return f"Grid({self.name} {self.__x_size__}x{self.__y_size__})"

    def add_entity(
        self,
        entity: WithSubclasses(Entity),
        time: int | float | None = None,
    ):
        """
        Adds an entity to the grid.

        Args:

        - entity (Entity): The entity to be added to the grid.
            - Must be an Entity or a subclass of Entity.
        - time (int|float|None): The time at which the entity should be added to the grid.
            - If None, the entity is added immediately.

        Returns:

        - Entity: The added entity.
        """
        entity.__assoc_grid__(self)
        self.__entities__[entity.id] = entity
        if time is None:
            entity.__place_on_grid__()
        else:
            self.add_event(
                time=time,
                object=entity,
                method="__place_on_grid__",
                kwargs={},
            )
        return entity

    def remove_entity(
        self, entity: WithSubclasses(Entity), time: int | float | None = None
    ):
        """
        Removes an entity from the grid.

        Args:

        - entity (Entity): The entity to be removed from the grid.
            - Must be an Entity or a subclass of Entity.
        - time (int|float|None): The time at which the entity should be removed from the grid.
            - If None, the entity is removed immediately.
        """
        if time is not None:
            self.add_event(
                time=time,
                object=self,
                method="remove_entity",
                kwargs={"entity": entity},
            )
            return entity
        if entity.id in self.__entities__:
            entity.__dissoc_grid__()
            self.__entities__.pop(entity.id, None)

    def add_event(
        self, time: int | float, object, method: str, kwargs: dict = dict()
    ) -> int:
        """
        Adds an event to the queue.
        This method schedules an event for a specific object at a specific time.
        Essentially, it allows you to schedule a method call on an object at a specific time.

        Args:

        - time (int|float): The time at which the event should occur.
        - object: The object on which the event will occur.
        - method (str): The name of the method to be called on the object.
        - kwargs (dict): The keyword arguments to be passed to the method.

        Returns:

        - int: The ID of the added event as generated by the queue.
        """
        return self.__queue__.add_event(
            time=time,
            event={"object": object, "method": method, "kwargs": kwargs},
        )

    def get_time(self) -> int | float:
        """
        Returns the current time of the grid.
        This method retrieves the current time from the queue.

        Returns:

        - int|float: The current time of the grid.
        """
        return self.__queue__.__time__

    def resolve_next_state(self) -> list[dict]:
        """
        Resolves the next state of the grid.
        This method processes the next event in the queue and updates the grid accordingly.

        Returns:

        - list[dict]: A list of events that were processed in this step.
            - Each event is a dictionary containing a time, id and event.
            - The event dictionary contains the object, method and kwargs.
                - id (int): The ID of the event as generated by the queue.
                - time (int|float): The time at which the event occurred.
                - event (dict): The event that was processed.
                    - object: The object on which the event occurred.
                    - method (str): The name of the method that was called.
                    - kwargs (dict): The keyword arguments that were passed to the method.
        """
        event_items = self.__queue__.get_next_events()
        for event_item in event_items:
            event = event_item.get("event")
            object = event.get("object")
            method = event.get("method")
            kwargs = event.get("kwargs")
            if object is None:
                continue
            getattr(object, method)(**kwargs)
        return event_items

    def add_exterior_walls(self) -> None:
        """
        Adds exterior walls to the grid.
        This method creates walls around the grid to prevent entities from moving outside the grid.
        """
        self.add_entity(
            StaticEntity(
                name="Left Wall",
                shape=[
                    [0, 0],
                    [0, self.__y_size__],
                    [1, self.__y_size__],
                    [1, 0],
                ],
                x_coord=0,
                y_coord=0,
            )
        )

        self.add_entity(
            StaticEntity(
                name="Right Wall",
                shape=[
                    [0, 0],
                    [0, self.__y_size__],
                    [-1, self.__y_size__],
                    [-1, 0],
                ],
                x_coord=self.__x_size__,
                y_coord=0,
            )
        )

        self.add_entity(
            StaticEntity(
                name="Top Wall",
                shape=[
                    [0, 0],
                    [self.__x_size__ - 2, 0],
                    [self.__x_size__ - 2, -1],
                    [0, -1],
                ],
                x_coord=1,
                y_coord=self.__y_size__,
            )
        )

        self.add_entity(
            StaticEntity(
                name="Bottom Wall",
                shape=[
                    [0, 0],
                    [self.__x_size__ - 2, 0],
                    [self.__x_size__ - 2, 1],
                    [0, 1],
                ],
                x_coord=1,
                y_coord=0,
            )
        )

    def simulate(self) -> None:
        """
        Runs the simulation for the grid.
        This method processes events in the queue until all events are resolved or the maximum time is reached.
        """
        next_state_events = True
        while next_state_events:
            next_state_events = self.resolve_next_state()



from fizgrid.grid import Grid
from fizgrid.entities import Entity, StaticEntity, GhostEntity
from fizgrid.utils import Shape

# Create a grid with exterior walls
grid = Grid(
    name="living_room",
    x_size=10,
    y_size=10,
    add_exterior_walls=True,
)

# Add some static entities
sofa = grid.add_entity(
    StaticEntity(
        name="sofa",
        shape=Shape.rectangle(x_len=2, y_len=1),
        x_coord=5,
        y_coord=5.5,
    )
)

# Add a dynamic entity
robot = grid.add_entity(
    Entity(
        name="robot",
        shape=Shape.rectangle(x_len=1, y_len=1),
        x_coord=5,
        y_coord=2,
    )
)

# Add a dynamic entity that can pass through other entities
ghost = grid.add_entity(
    GhostEntity(
        name="ghost",
        shape=Shape.rectangle(x_len=1, y_len=1),
        x_coord=5,
        y_coord=2,
    )
)

# Attempt to move the robot through the sofa
# Starting at time=0, the robot will try to move to (5,8) over 12 seconds
# This will bump into the sofa and stop
robot.add_route(time=0, waypoints=[(5, 8, 12)])


# Attempt to move the ghost through the sofa
# Starting at time=0, the ghost will try to move to (5,8) over 12 seconds
# This will not bump into the sofa and complete its route
ghost.add_route(time=0, waypoints=[(5, 8, 12)])

# Simulate the grid
grid.simulate()

# Show the history of the robot
# print(robot.history)
# [{'x': 5, 'y': 2, 't': 0, 'c': False}, {'x': 5.0, 'y': 4.5, 't': 5.0, 'c': True}]
# This means that the robot started at (5,2) at time=0 and moved to (5,4.5) at time=5 where it collided with the sofa
# Since there is no additional logic to handle the collision, the robot just stopped moving at time 5 and waited for the end of the simulation


# Show the history of the ghost
# print(ghost.history)
# [{'x': 5, 'y': 2, 't': 0, 'c': False}, {'x': 5, 'y': 8, 't': 12, 'c': False}]
# This means that the ghost started at (5,2) at time=0 and moved to (5,8) at time=12 without any collisions

success = True
if robot.history[-1]["y"] != 4.5 and robot.history[-1]["c"] != True:
    success = False
if ghost.history[-1]["y"] != 8 and ghost.history[-1]["c"] != False:
    success = False

if success:
    print("test_07.py: passed")
else:
    print("test_07.py: failed")
